﻿module proxy;

private import
	std.thread, std.string, std.stream, std.gc,
	socket, http;

alias char[] function(char[], char[]) Html_Filter;

char[] take_first_line(inout char[] text)
{
	char[] result;
	int i = find(text, '\n');
	if(i >= 0){
		result = text[0..i + 1];
		text = text[i + 1..text.length];
	}else{
		result = text;
		text = ``;
	}
	return result;
}

bit take_front(inout char[] text, in char[] s)
{
	bit result = false;
	if(text.length >= s.length && text[0..s.length] == s){
		text = text[s.length..text.length];
		result = true;
	}
	return result;
}

/+
char[] escape(char[] s)
{
	char[] result = "";
	for(int i = 0; i < s.length; ++i){
		if(s[i] < ' ' && s[i] != 10 && s[i] != 13){
			result ~= "?";
		}else{
			result ~= s[i];
		}
	}
	return result;
}
+/

class Response_Thread : Thread
{
	static int thread_count = 0;
	static Object thread_count_sync;

	static this()
	{
		thread_count_sync = new Object;
	}

	Response_Socket response;
	Html_Filter html_filter;

	const char[] cl = "Content-Length:";
	const char[] ct = "Content-Type:";

	override int run()
	{
		synchronized(thread_count_sync){
			++thread_count;
			stdout.printf(`---- #%d start (%d)`\n, id, thread_count);
		}
	
		try{
			ubyte[4096] buffer;
			int received;

			/+ char[] request = cast(char[])response.receive(int.max); +/
			char[] request = cast(char[])response.receive_all();
			
			stdout.printf(`---- #%d from browser`\n`%.*s`\n, id, request);
			
			char[] source = request;
			char[] uri = take_first_line(source);
			char[] cmd;

			if(take_front(uri, "GET ")){
				cmd = "GET ";
			}else if(take_front(uri, "HEAD ")){
				cmd = "HEAD ";
			}else if(take_front(uri, "POST ")){
				cmd = "POST ";
			}else{
				throw new Exception(`未対応リクエスト: ` ~ uri);
			}
			
			while(uri[uri.length - 1] <= '\x20'){
				uri = uri[0..uri.length - 1];
			}
			if(cmp(uri[uri.length - 9..uri.length], " HTTP/1.0") == 0){
				uri = uri[0..uri.length - 9];
			}
			
			ushort port;
			char[] server_name;
			char[] filename;
			disassemble_uri(uri, server_name, port, filename);
			IP_Address ip = to_ip(server_name);
			
			char[] request2 = cmd ~ filename ~ ` HTTP/1.0`\r\n;
			
			int source_length = 0;
			while(source.length > 0){
				char[] line = take_first_line(source);
				if(line == \n || line == \r\n){
					request2 ~= line;
					break;
				}else if(line == `Proxy-Connection: Keep-Alive`\r\n){
					line = ``;
				}else if(line.length >= cl.length && icmp(line[0..cl.length], cl) == 0){
					char[] n = line[cl.length..line.length];
					while(n.length > 0 && n[0] <= ' '){ 
						n = n[1..n.length]; 
					}
					source_length = scan_int(n);
				}
				request2 ~= line;
			}
			
			while(source.length < source_length){
				received = response.receive(buffer);
				if(source.length == 0){
					source = cast(char[])buffer[0..received].dup;
				}else{
					source ~= cast(char[])buffer[0..received];
				}
				if(source.length > source_length){
					source.length = source_length;
				}
			}
			
			stdout.printf("---- #%d to server\n%.*s\n", id, request2);
			if(source.length > 0){
				stdout.printf("---- #%d to server %d byte.\n", id, source.length);
			}
			
			Client_Socket client = new Client_Socket(ip, port);
			try{
				client.send(cast(ubyte[])request2);
				if(source.length > 0){
					//がさっと送るとH"が耐えられないので1kづつ小分けに…
					while(source.length > 0){
						int n = 1024;
						if(n > source.length){
							n = source.length;
						}
						client.send(cast(ubyte[])source[0..n]);
						source = source[n..source.length];
					}
				}
				stdout.printf(`---- #%d receiving from server`\n, id);
				
				/+char[] contents = cast(char[])(client.receive(int.max));+/
				received = client.receive(buffer);
				char[] contents = cast(char[])buffer[0..received].dup;

				//stdout.printf("%.*s\n", escape(contents));
				//stdout.printf("%d received %d byte\n", id, contents.length);
	
				int data_length = 0;
				bit is_html = false;
				char[] header = "";
				for(; contents.length > 0;){
					char[] line = take_first_line(contents);
					if(line == "\r\n" || line == "\n"){
						header ~= line;
						break;
					}
					if(line.length >= cl.length && icmp(line[0..cl.length], cl) == 0){
						char[] n = line[cl.length..line.length];
						while(n.length > 0 && n[0] <= ' '){ 
							n = n[1..n.length]; 
						}
						data_length = scan_int(n);
					}else if(line.length >= ct.length && icmp(line[0..ct.length], ct) == 0){
						char[] n = line[ct.length..line.length];
						while(n.length > 0 && n[0] <= ' '){ 
							n = n[1..n.length]; 
						}
						
						int sc = find(n, ';');
						if(sc >= 0){
							n = n[0..sc];
						}
						
						while(n.length > 0 && n[n.length - 1] <= ' '){ 
							n = n[0..n.length - 1]; 
						}
						//stdout.printf(`%d's contents-type "%.*s"`\n, id, n);
						if(icmp(n, "text/html") == 0){
							is_html = true;
						}
					}
					
					header ~= line;
				}
				
				stdout.writeLine(header);
				if(is_html){
					stdout.printf("** #%d receiving html **\n", id);
				}
				if(data_length > 0){
					stdout.printf("#%d content-length: %d\n", id, data_length);
				}
				stdout.printf("#%d received %d byte\n", id, contents.length);

				response.send(cast(ubyte[])header);
				
				if(!is_html){
					response.send(cast(ubyte[])contents);
				}else if(contents.length > 0){
					contents = contents.dup;
				}
				
				int zero_count = 0;
				while(received > 0 || contents.length < data_length){
					received = client.receive(buffer);
					if(is_html){
						if(received > 0){
							if(contents.length == 0){
								contents = cast(char[])buffer[0..received].dup;
							}else{
								char[] old = contents;
								contents ~= cast(char[])buffer[0..received];
								if(cast(char*)contents != cast(char*)old){
									delete old;
								}
							}
						}
						stdout.printf("#%d received %d byte, total %d \n", id, received, contents.length);
					}else{
						response.send(cast(ubyte[])(buffer[0..received]));
						stdout.printf("#%d received %d byte\n", id, received);
					}
					//100回0が続けばあきらめる(受信エラーにならない時が時々ある)
					if(received == 0){
						++zero_count;
						if(zero_count == 100) break;
					}else{
						zero_count = 0;
					}
				}
				
				if(is_html){
					contents = html_filter(uri, contents);
					response.send(cast(ubyte[])contents);
				}
				
			}finally{
				stdout.printf(`---- #%d close`\n, id);
				client.close();
			}
		}catch(Exception e){
			stderr.printf(`error in response thread #%d %.*s (%.*s)`\r\n, id, e.toString(), e.classinfo.name);
			char[] error = 
				`HTTP/1.1 500 Proxy Error`\r\n
				`Connection: close`\r\n
				\r\n
				`﻿` ~ e.toString();
			try{
				response.send(cast(ubyte[])error);
			}catch{
				//何もできない...
			}
		}
		
		response.close();
		
		synchronized(thread_count_sync){
			--thread_count;
			stdout.printf("---- #%d end (%d) \n", id, thread_count);
			if(thread_count == 0){
				std.gc.fullCollect();
				std.gc.minimize();
			}
		}

		return 0;
	}
}

class Proxy_Thread : Thread
{
	Html_Filter html_filter;
	bit terminating;
	
	override int run()
	{
		try{
			Server_Socket server = new Server_Socket(8080);
			for(;;){
				Response_Socket response = server.accept();
				if(terminating) break;
				Response_Thread thread = new Response_Thread();
				thread.response = response;
				thread.html_filter = html_filter;
				thread.start();
			}
		}catch(Exception e){
			stdout.printf("error in server thread (%.*s): %.*s\n", e.classinfo.name, e.toString);
		}
		return 0;
	}
}
