module socket;

private import
	std.string, std.stdint, win32.winsock;

typedef uint32_t IP_Address;

static this()
{
	/* WinSockの初期化を行う */
	WORD version_requested;
	WSADATA data;
	version_requested = MAKEWORD(1, 1);
	int error_status = WSAStartup(version_requested, &data);
	if(error_status != 0){
		throw new Exception("WinSockの初期化失敗");
	}
}

static ~this()
{
	WSACleanup();
}

IP_Address to_ip(char[] server_name)
body {
	IP_Address result;
	char* name = toStringz(server_name);
	result = cast(IP_Address)inet_addr(name); /* A.B.C.D */
	if(result == INADDR_NONE){
		hostent* server_hostent = gethostbyname(name); /* www.xxx.net */
		if(server_hostent == NULL){
			throw new Exception("ホストアドレス取得失敗");
		}
		result = *(cast(IP_Address*)(server_hostent.h_addr_list[0]));
	}
	return result;
}

class Socket : Object
{
	this(SOCKET _socket)
	{
		socket = _socket;
		//received_finished = false;
	}

	~this()
	{
		close();
	}
	
	void close()
	{
		if(socket != INVALID_SOCKET){
			destroy_socket(socket);
			socket = INVALID_SOCKET;
		}
	}
	
	//bit received_finished;
	
protected:

	void send(ubyte[] buffer)
	{
		send(socket, buffer);
	}

/+
	ubyte[] receive(int max_length)
	{
		return receive(socket, max_length, received_finished);
	}
+/
	int receive(ubyte[] buffer)
	{
		return receive(socket, buffer);
	}

	ubyte[] receive_all()
	{
		ubyte[4096] buffer;
		ubyte[] result = null;
		int r;
		while((r = receive(buffer)) > 0){
			if(result.length == 0){
				result = buffer[0..r].dup;
			}else{
				assert(result.length > 0);
				assert(r > 0);
				ubyte[] old = result;
				result ~= buffer[0..r];
				delete old;
			}
			if(r < buffer.length) break;
		}
		return result;
	}

private:
	SOCKET socket;

	static SOCKET create_socket()
	{
		SOCKET result = win32.winsock.socket(AF_INET, SOCK_STREAM, 0);
		if(result == INVALID_SOCKET){
			throw new Exception("ソケット作成失敗");
		}
		return result;
	}

	static void destroy_socket(SOCKET socket)
	{
		shutdown(socket, 2);
		closesocket(socket);
	}

	static void send(SOCKET socket, ubyte[] buffer)
	{
		//printf("%.*s\n", buffer);
		if(buffer.length > 0){
			if(win32.winsock.send(socket, cast(char*)(&buffer[0]), buffer.length, 0) == SOCKET_ERROR){
				throw new Exception("送信失敗" ~ std.string.toString(WSAGetLastError()));
			}
		}
	}

/+
	static ubyte[] receive(SOCKET socket, int max_length, out bit fin)
	{
		ubyte[] result = null;
		fin = false;
		while(result.length < max_length){
			ubyte[256] buffer;
			int bs = max_length - result.length;
			if(bs > buffer.length){
				bs = buffer.length;
			}
			//printf(`begin receive %d `\n, bs);
			int r = recv(socket, cast(char*)(&buffer[0]), bs, 0);
			//printf(`end receive`\n);
			if(r == SOCKET_ERROR){
				throw new Exception("受信失敗");
			}else if(r == 0){
				fin = true;
				break;
			}
			//printf("<%d",r);
			result ~= buffer[0..r];
			//printf(">");
			if(r < bs){
				break;
			}
		}
		//printf("return from receive"\n);
		return result;
	}
+/
	static int receive(SOCKET socket, ubyte[] buffer)
	{
		int result = recv(socket, cast(char*)(&buffer[0]), buffer.length, 0);
		if(result == SOCKET_ERROR){
				throw new Exception("受信失敗" ~ std.string.toString(WSAGetLastError()));
		}
		return result;
	}
}

class Client_Socket : Socket
{
	this(IP_Address server_ip, ushort port){
		super(create_socket);
		connect(socket, server_ip, port);
	}

	override void send(ubyte[] buffer)
	{
		super.send(buffer);
	}

/+
	override ubyte[] receive(int max_length)
	{
		return super.receive(max_length);
	}
+/
	override int receive(ubyte[] buffer)
	{
		return super.receive(buffer);
	}

	override ubyte[] receive_all()
	{
		return super.receive_all();
	}

private:
	static void connect(SOCKET socket, uint32_t ip_address, ushort port)
	body {
		sockaddr_in server_sock_addr;
		std.string.memset(&server_sock_addr, 0, server_sock_addr.sizeof);
		server_sock_addr.sin_family = AF_INET;
		server_sock_addr.sin_addr.S_addr = ip_address;
		server_sock_addr.sin_port = htons(port);
		if(win32.winsock.connect(socket, 
			cast(sockaddr*)(&server_sock_addr), 
			server_sock_addr.sizeof) == SOCKET_ERROR)
		{
			throw new Exception("サーバへの接続失敗");
		}
	}
}

class Server_Socket : Socket
{
	this(ushort port){
		super(create_socket);
		bind(socket, port);
		listen(socket, SOMAXCONN);
	}

	Response_Socket accept()
	{
		return new Response_Socket(accept(socket));
	}

private:

	void bind(SOCKET socket, ushort port)
	{
		sockaddr_in reader_addr;
		std.string.memset(&reader_addr, 0, reader_addr.sizeof);
		reader_addr.sin_family = AF_INET;
		reader_addr.sin_addr.S_addr = htonl(INADDR_ANY);
		reader_addr.sin_port = htons(port);
		if(win32.winsock.bind(socket, cast(sockaddr*)&reader_addr, reader_addr.sizeof) == SOCKET_ERROR){
			throw new Exception("bind失敗");
		}
	}
  
	void listen(SOCKET socket, int max)
	{
		if(win32.winsock.listen(socket, max) == SOCKET_ERROR){
			throw new Exception("listen失敗");
		}
	}
	
	SOCKET accept(SOCKET socket)
	{
		SOCKET result = win32.winsock.accept(socket, null, null);
		if(result == INVALID_SOCKET){
			throw new Exception("accept失敗");
		}
		return result;
	}
}

class Response_Socket : Socket
{
	this(SOCKET _socket)
	{
		super(_socket);
	}
	
	override void send(ubyte[] buffer)
	{
		super.send(buffer);
	}

/+
	override ubyte[] receive(int max_length)
	{
		return super.receive(max_length);
	}
+/
	override int receive(ubyte[] buffer)
	{
		return super.receive(buffer);
	}

	override ubyte[] receive_all()
	{
		return super.receive_all();
	}
}

